@@ -1,5 +1,6 @@
 /* Dead store elimination
-   Copyright (C) 2004, 2005, 2006, 2007 Free Software Foundation, Inc.
+   Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009
+   Free Software Foundation, Inc.
 
 This file is part of GCC.
 
@@ -33,6 +34,7 @@ along with GCC; see the file COPYING3.  If not see
 #include "tree-dump.h"
 #include "domwalk.h"
 #include "flags.h"
+#include "langhooks.h"
 
 /* This file implements dead store elimination.
 
@@ -62,7 +64,7 @@ along with GCC; see the file COPYING3.  If not see
    relationship between dead store and redundant load elimination.  In
    fact, they are the same transformation applied to different views of
    the CFG.  */
-   
+
 
 struct dse_global_data
 {
@@ -96,24 +98,21 @@ static void dse_initialize_block_local_data (struct dom_walk_data *,
 					     bool);
 static void dse_optimize_stmt (struct dom_walk_data *,
 			       basic_block,
-			       block_stmt_iterator);
+			       gimple_stmt_iterator);
 static void dse_record_phis (struct dom_walk_data *, basic_block);
 static void dse_finalize_block (struct dom_walk_data *, basic_block);
 static void record_voperand_set (bitmap, bitmap *, unsigned int);
 
-static unsigned max_stmt_uid;	/* Maximal uid of a statement.  Uids to phi
-				   nodes are assigned using the versions of
-				   ssa names they define.  */
-
 /* Returns uid of statement STMT.  */
 
 static unsigned
-get_stmt_uid (tree stmt)
+get_stmt_uid (gimple stmt)
 {
-  if (TREE_CODE (stmt) == PHI_NODE)
-    return SSA_NAME_VERSION (PHI_RESULT (stmt)) + max_stmt_uid;
+  if (gimple_code (stmt) == GIMPLE_PHI)
+    return SSA_NAME_VERSION (gimple_phi_result (stmt))
+           + gimple_stmt_max_uid (cfun);
 
-  return stmt_ann (stmt)->uid;
+  return gimple_uid (stmt);
 }
 
 /* Set bit UID in bitmaps GLOBAL and *LOCAL, creating *LOCAL as needed.  */
@@ -165,7 +164,7 @@ memory_ssa_name_same (tree *expr_p, int *walk_subtrees ATTRIBUTE_UNUSED,
 {
   struct address_walk_data *walk_data = (struct address_walk_data *) data;
   tree expr = *expr_p;
-  tree def_stmt;
+  gimple def_stmt;
   basic_block def_bb;
 
   if (TREE_CODE (expr) != SSA_NAME)
@@ -177,7 +176,7 @@ memory_ssa_name_same (tree *expr_p, int *walk_subtrees ATTRIBUTE_UNUSED,
     return NULL_TREE;
 
   def_stmt = SSA_NAME_DEF_STMT (expr);
-  def_bb = bb_for_stmt (def_stmt);
+  def_bb = gimple_bb (def_stmt);
 
   /* DEF_STMT must dominate both stores.  So if it is in the same
      basic block as one, it does not post-dominate that store.  */
@@ -188,7 +187,7 @@ memory_ssa_name_same (tree *expr_p, int *walk_subtrees ATTRIBUTE_UNUSED,
 	  || !dominated_by_p (CDI_POST_DOMINATORS, walk_data->store2_bb,
 			      def_bb))
 	/* Return non-NULL to stop the walk.  */
-	return def_stmt;
+	return *expr_p;
     }
 
   return NULL_TREE;
@@ -198,34 +197,34 @@ memory_ssa_name_same (tree *expr_p, int *walk_subtrees ATTRIBUTE_UNUSED,
    might be modified after STORE1, before control reaches STORE2.  */
 
 static bool
-memory_address_same (tree store1, tree store2)
+memory_address_same (gimple store1, gimple store2)
 {
   struct address_walk_data walk_data;
 
-  walk_data.store1_bb = bb_for_stmt (store1);
-  walk_data.store2_bb = bb_for_stmt (store2);
+  walk_data.store1_bb = gimple_bb (store1);
+  walk_data.store2_bb = gimple_bb (store2);
 
-  return (walk_tree (&GIMPLE_STMT_OPERAND (store1, 0), memory_ssa_name_same,
+  return (walk_tree (gimple_assign_lhs_ptr (store1), memory_ssa_name_same,
 		     &walk_data, NULL)
 	  == NULL);
 }
 
 /* Return true if there is a stmt that kills the lhs of STMT and is in the
-   virtual def-use chain of STMT without a use inbetween the kill and STMT.
+   virtual def-use chain of STMT without a use in between the kill and STMT.
    Returns false if no such stmt is found.
    *FIRST_USE_P is set to the first use of the single virtual def of
    STMT.  *USE_P is set to the vop killed by *USE_STMT.  */
 
 static bool
-get_kill_of_stmt_lhs (tree stmt,
+get_kill_of_stmt_lhs (gimple stmt,
 		      use_operand_p * first_use_p,
- 		      use_operand_p * use_p, tree * use_stmt)
+ 		      use_operand_p * use_p, gimple * use_stmt)
 {
   tree lhs;
 
-  gcc_assert (TREE_CODE (stmt) == GIMPLE_MODIFY_STMT);
+  gcc_assert (is_gimple_assign (stmt));
 
-  lhs = GIMPLE_STMT_OPERAND (stmt, 0);
+  lhs = gimple_assign_lhs (stmt);
 
   /* We now walk the chain of single uses of the single VDEFs.
      We succeeded finding a kill if the lhs of the use stmt is
@@ -234,7 +233,7 @@ get_kill_of_stmt_lhs (tree stmt,
      the stmt.  */
   do
     {
-      tree use_lhs, use_rhs;
+      tree use_lhs;
       def_operand_p def_p;
 
       /* The stmt must have a single VDEF.  */
@@ -248,17 +247,14 @@ get_kill_of_stmt_lhs (tree stmt,
       first_use_p = use_p;
 
       /* If there are possible hidden uses, give up.  */
-      if (TREE_CODE (stmt) != GIMPLE_MODIFY_STMT)
-	return false;
-      use_rhs = GIMPLE_STMT_OPERAND (stmt, 1);
-      if (TREE_CODE (use_rhs) == CALL_EXPR
-	  || (!is_gimple_min_invariant (use_rhs)
-	      && TREE_CODE (use_rhs) != SSA_NAME))
+      if (!gimple_assign_single_p (stmt)
+	  || (TREE_CODE (gimple_assign_rhs1 (stmt)) != SSA_NAME
+	      && !is_gimple_min_invariant (gimple_assign_rhs1 (stmt))))
 	return false;
 
       /* If the use stmts lhs matches the original lhs we have
 	 found the kill, otherwise continue walking.  */
-      use_lhs = GIMPLE_STMT_OPERAND (stmt, 0);
+      use_lhs = gimple_assign_lhs (stmt);
       if (operand_equal_p (use_lhs, lhs, 0))
 	{
 	  *use_stmt = stmt;
@@ -269,16 +265,16 @@ get_kill_of_stmt_lhs (tree stmt,
 }
 
 /* A helper of dse_optimize_stmt.
-   Given a GIMPLE_MODIFY_STMT in STMT, check that each VDEF has one
+   Given a GIMPLE_ASSIGN in STMT, check that each VDEF has one
    use, and that one use is another VDEF clobbering the first one.
 
    Return TRUE if the above conditions are met, otherwise FALSE.  */
 
 static bool
-dse_possible_dead_store_p (tree stmt,
+dse_possible_dead_store_p (gimple stmt,
 			   use_operand_p *first_use_p,
 			   use_operand_p *use_p,
-			   tree *use_stmt,
+			   gimple *use_stmt,
 			   struct dse_global_data *dse_gd,
 			   struct dse_block_local_data *bd)
 {
@@ -286,9 +282,9 @@ dse_possible_dead_store_p (tree stmt,
   bool fail = false;
   def_operand_p var1;
   vuse_vec_p vv;
-  tree defvar = NULL_TREE, temp;
+  tree defvar = NULL_TREE;
   tree prev_defvar = NULL_TREE;
-  stmt_ann_t ann = stmt_ann (stmt);
+  gimple temp;
 
   /* We want to verify that each virtual definition in STMT has
      precisely one use and that all the virtual definitions are
@@ -313,6 +309,14 @@ dse_possible_dead_store_p (tree stmt,
       gcc_assert (*use_p != NULL_USE_OPERAND_P);
       *first_use_p = *use_p;
 
+      /* ???  If we hit a GIMPLE_PHI we could skip to the PHI_RESULT uses.
+	 Don't bother to do that for now.  */
+      if (gimple_code (temp) == GIMPLE_PHI)
+	{
+	  fail = true;
+	  break;
+	}
+
       /* In the case of memory partitions, we may get:
 
 	   # MPT.764_162 = VDEF <MPT.764_161(D)>
@@ -322,10 +326,10 @@ dse_possible_dead_store_p (tree stmt,
 
 	   So we must make sure we're talking about the same LHS.
       */
-      if (TREE_CODE (temp) == GIMPLE_MODIFY_STMT)
+      if (is_gimple_assign (temp))
 	{
-	  tree base1 = get_base_address (GIMPLE_STMT_OPERAND (stmt, 0));
-	  tree base2 =  get_base_address (GIMPLE_STMT_OPERAND (temp, 0));
+	  tree base1 = get_base_address (gimple_assign_lhs (stmt));
+	  tree base2 = get_base_address (gimple_assign_lhs (temp));
 
 	  while (base1 && INDIRECT_REF_P (base1))
 	    base1 = TREE_OPERAND (base1, 0);
@@ -356,33 +360,10 @@ dse_possible_dead_store_p (tree stmt,
 
   if (fail)
     {
-      record_voperand_set (dse_gd->stores, &bd->stores, ann->uid);
+      record_voperand_set (dse_gd->stores, &bd->stores, gimple_uid (stmt));
       return false;
     }
 
-  /* Skip through any PHI nodes we have already seen if the PHI
-     represents the only use of this store.
-
-     Note this does not handle the case where the store has
-     multiple VDEFs which all reach a set of PHI nodes in the same block.  */
-  while (*use_p != NULL_USE_OPERAND_P
-	 && TREE_CODE (*use_stmt) == PHI_NODE
-	 && bitmap_bit_p (dse_gd->stores, get_stmt_uid (*use_stmt)))
-    {
-      /* A PHI node can both define and use the same SSA_NAME if
-	 the PHI is at the top of a loop and the PHI_RESULT is
-	 a loop invariant and copies have not been fully propagated.
-
-	 The safe thing to do is exit assuming no optimization is
-	 possible.  */
-      if (SSA_NAME_DEF_STMT (PHI_RESULT (*use_stmt)) == *use_stmt)
-	return false;
-
-      /* Skip past this PHI and loop again in case we had a PHI
-	 chain.  */
-      single_imm_use (PHI_RESULT (*use_stmt), use_p, use_stmt);
-    }
-
   return true;
 }
 
@@ -401,36 +382,35 @@ dse_possible_dead_store_p (tree stmt,
 static void
 dse_optimize_stmt (struct dom_walk_data *walk_data,
 		   basic_block bb ATTRIBUTE_UNUSED,
-		   block_stmt_iterator bsi)
+		   gimple_stmt_iterator gsi)
 {
   struct dse_block_local_data *bd
     = (struct dse_block_local_data *)
 	VEC_last (void_p, walk_data->block_data_stack);
   struct dse_global_data *dse_gd
     = (struct dse_global_data *) walk_data->global_data;
-  tree stmt = bsi_stmt (bsi);
-  stmt_ann_t ann = stmt_ann (stmt);
+  gimple stmt = gsi_stmt (gsi);
 
   /* If this statement has no virtual defs, then there is nothing
      to do.  */
   if (ZERO_SSA_OPERANDS (stmt, SSA_OP_VDEF))
     return;
 
-  /* We know we have virtual definitions.  If this is a GIMPLE_MODIFY_STMT
+  /* We know we have virtual definitions.  If this is a GIMPLE_ASSIGN
      that's not also a function call, then record it into our table.  */
-  if (get_call_expr_in (stmt))
+  if (is_gimple_call (stmt) && gimple_call_fndecl (stmt))
     return;
 
-  if (ann->has_volatile_ops)
+  if (gimple_has_volatile_ops (stmt))
     return;
 
-  if (TREE_CODE (stmt) == GIMPLE_MODIFY_STMT)
+  if (is_gimple_assign (stmt))
     {
       use_operand_p first_use_p = NULL_USE_OPERAND_P;
       use_operand_p use_p = NULL;
-      tree use_stmt;
+      gimple use_stmt;
 
-      if (!dse_possible_dead_store_p (stmt, &first_use_p, &use_p, &use_stmt,
+      if (!dse_possible_dead_store_p (stmt, &first_use_p, &use_p, &use_stmt, 
 				      dse_gd, bd))
 	return;
 
@@ -440,8 +420,8 @@ dse_optimize_stmt (struct dom_walk_data *walk_data,
 	 SSA-form variables in the address will have the same values.  */
       if (use_p != NULL_USE_OPERAND_P
           && bitmap_bit_p (dse_gd->stores, get_stmt_uid (use_stmt))
-          && !operand_equal_p (GIMPLE_STMT_OPERAND (stmt, 0),
-                               GIMPLE_STMT_OPERAND (use_stmt, 0), 0)
+          && !operand_equal_p (gimple_assign_lhs (stmt),
+                               gimple_assign_lhs (use_stmt), 0)
           && memory_address_same (stmt, use_stmt))
         {
           /* If we have precisely one immediate use at this point, but
@@ -450,7 +430,8 @@ dse_optimize_stmt (struct dom_walk_data *walk_data,
              memory location.  */
           if (!get_kill_of_stmt_lhs (stmt, &first_use_p, &use_p, &use_stmt))
             {
-              record_voperand_set (dse_gd->stores, &bd->stores, ann->uid);
+              record_voperand_set (dse_gd->stores, &bd->stores, 
+				   gimple_uid (stmt));
               return;
             }
         }
@@ -461,8 +442,8 @@ dse_optimize_stmt (struct dom_walk_data *walk_data,
 	 memory location, then we may have found redundant store.  */
       if (use_p != NULL_USE_OPERAND_P
 	  && bitmap_bit_p (dse_gd->stores, get_stmt_uid (use_stmt))
-	  && operand_equal_p (GIMPLE_STMT_OPERAND (stmt, 0),
-			      GIMPLE_STMT_OPERAND (use_stmt, 0), 0)
+	  && operand_equal_p (gimple_assign_lhs (stmt),
+			      gimple_assign_lhs (use_stmt), 0)
 	  && memory_address_same (stmt, use_stmt))
 	{
 	  ssa_op_iter op_iter;
@@ -470,30 +451,30 @@ dse_optimize_stmt (struct dom_walk_data *walk_data,
 	  vuse_vec_p vv;
 	  tree stmt_lhs;
 
-	  if (LOADED_SYMS (use_stmt))
+	  /* If use_stmt is or might be a nop assignment, e.g. for
+	     struct { ... } S a, b, *p; ...
+	     b = a; b = b;
+	     or
+	     b = a; b = *p; where p might be &b,
+	     or
+	     *p = a; *p = b; where p might be &b,
+	     or
+	     *p = *u; *p = *v; where p might be v, then USE_STMT
+	     acts as a use as well as definition, so store in STMT
+	     is not dead.  */
+	  if (gimple_loaded_syms (use_stmt)
+	      && bitmap_intersect_p (gimple_loaded_syms (use_stmt),
+				     gimple_stored_syms (use_stmt)))
 	    {
-	      tree use_base
-		= get_base_address (GIMPLE_STMT_OPERAND (use_stmt, 0));
-	      /* If use_stmt is or might be a nop assignment, e.g. for
-		 struct { ... } S a, b, *p; ...
-		 b = a; b = b;
-		 or
-		 b = a; b = *p; where p might be &b, then USE_STMT
-		 acts as a use as well as definition, so store in STMT
-		 is not dead.  */
-	      if (TREE_CODE (use_base) == VAR_DECL
-		  && bitmap_bit_p (LOADED_SYMS (use_stmt),
-				   DECL_UID (use_base)))
-		{
-		  record_voperand_set (dse_gd->stores, &bd->stores, ann->uid);
-		  return;
-		}
+              record_voperand_set (dse_gd->stores, &bd->stores, 
+				   gimple_uid (stmt));
+	      return;
 	    }
 
 	  if (dump_file && (dump_flags & TDF_DETAILS))
             {
               fprintf (dump_file, "  Deleted dead store '");
-              print_generic_expr (dump_file, bsi_stmt (bsi), dump_flags);
+              print_gimple_stmt (dump_file, gsi_stmt (gsi), dump_flags, 0);
               fprintf (dump_file, "'\n");
             }
 
@@ -501,7 +482,8 @@ dse_optimize_stmt (struct dom_walk_data *walk_data,
 	  stmt_lhs = USE_FROM_PTR (first_use_p);
 	  FOR_EACH_SSA_VDEF_OPERAND (var1, vv, stmt, op_iter)
 	    {
-	      tree usevar, temp;
+	      tree usevar;
+	      gimple temp;
 
 	      single_imm_use (DEF_FROM_PTR (var1), &use_p, &temp);
 	      gcc_assert (VUSE_VECT_NUM_ELEM (*vv) == 1);
@@ -514,14 +496,14 @@ dse_optimize_stmt (struct dom_walk_data *walk_data,
 	    }
 
 	  /* Remove the dead store.  */
-	  bsi_remove (&bsi, true);
+	  gsi_remove (&gsi, true);
 
 	  /* And release any SSA_NAMEs set in this statement back to the
 	     SSA_NAME manager.  */
 	  release_defs (stmt);
 	}
 
-      record_voperand_set (dse_gd->stores, &bd->stores, ann->uid);
+      record_voperand_set (dse_gd->stores, &bd->stores, gimple_uid (stmt));
     }
 }
 
@@ -535,13 +517,15 @@ dse_record_phis (struct dom_walk_data *walk_data, basic_block bb)
 	VEC_last (void_p, walk_data->block_data_stack);
   struct dse_global_data *dse_gd
     = (struct dse_global_data *) walk_data->global_data;
-  tree phi;
+  gimple phi;
+  gimple_stmt_iterator gsi;
 
-  for (phi = phi_nodes (bb); phi; phi = PHI_CHAIN (phi))
-    if (!is_gimple_reg (PHI_RESULT (phi)))
-      record_voperand_set (dse_gd->stores,
-			   &bd->stores,
-			   get_stmt_uid (phi));
+  for (gsi = gsi_start_phis (bb); !gsi_end_p (gsi); gsi_next (&gsi))
+    {
+      phi = gsi_stmt (gsi);
+      if (!is_gimple_reg (gimple_phi_result (phi)))
+	record_voperand_set (dse_gd->stores, &bd->stores, get_stmt_uid (phi));
+    }
 }
 
 static void
@@ -572,18 +556,8 @@ tree_ssa_dse (void)
 {
   struct dom_walk_data walk_data;
   struct dse_global_data dse_gd;
-  basic_block bb;
 
-  /* Create a UID for each statement in the function.  Ordering of the
-     UIDs is not important for this pass.  */
-  max_stmt_uid = 0;
-  FOR_EACH_BB (bb)
-    {
-      block_stmt_iterator bsi;
-
-      for (bsi = bsi_start (bb); !bsi_end_p (bsi); bsi_next (&bsi))
-	stmt_ann (bsi_stmt (bsi))->uid = max_stmt_uid++;
-    }
+  renumber_gimple_stmt_uids ();
 
   /* We might consider making this a property of each pass so that it
      can be [re]computed on an as-needed basis.  Particularly since
@@ -633,7 +607,10 @@ gate_dse (void)
   return flag_tree_dse != 0;
 }
 
-struct tree_opt_pass pass_dse = {
+struct gimple_opt_pass pass_dse = 
+{
+ {
+  GIMPLE_PASS,
   "dse",			/* name */
   gate_dse,			/* gate */
   tree_ssa_dse,			/* execute */
@@ -649,8 +626,8 @@ struct tree_opt_pass pass_dse = {
   0,				/* todo_flags_start */
   TODO_dump_func
     | TODO_ggc_collect
-    | TODO_verify_ssa,		/* todo_flags_finish */
-  0				/* letter */
+    | TODO_verify_ssa		/* todo_flags_finish */
+ }
 };
 
 /* A very simple dead store pass eliminating write only local variables.
@@ -660,7 +637,7 @@ struct tree_opt_pass pass_dse = {
 static unsigned int
 execute_simple_dse (void)
 {
-  block_stmt_iterator bsi;
+  gimple_stmt_iterator gsi;
   basic_block bb;
   bitmap variables_loaded = BITMAP_ALLOC (NULL);
   unsigned int todo = 0;
@@ -668,31 +645,51 @@ execute_simple_dse (void)
   /* Collect into VARIABLES LOADED all variables that are read in function
      body.  */
   FOR_EACH_BB (bb)
-    for (bsi = bsi_start (bb); !bsi_end_p (bsi); bsi_next (&bsi))
-      if (LOADED_SYMS (bsi_stmt (bsi)))
+    for (gsi = gsi_start_bb (bb); !gsi_end_p (gsi); gsi_next (&gsi))
+
+      if (gimple_loaded_syms (gsi_stmt (gsi)))
 	bitmap_ior_into (variables_loaded,
-			 LOADED_SYMS (bsi_stmt (bsi)));
+			 gimple_loaded_syms (gsi_stmt (gsi)));
 
   /* Look for statements writing into the write only variables.
      And try to remove them.  */
 
   FOR_EACH_BB (bb)
-    for (bsi = bsi_start (bb); !bsi_end_p (bsi);)
+    for (gsi = gsi_start_bb (bb); !gsi_end_p (gsi);)
       {
-	tree stmt = bsi_stmt (bsi), op;
+	gimple stmt = gsi_stmt (gsi);
+        tree op;
 	bool removed = false;
         ssa_op_iter iter;
+	tree size;
 
-	if (STORED_SYMS (stmt) && TREE_CODE (stmt) == GIMPLE_MODIFY_STMT
-	    && TREE_CODE (stmt) != RETURN_EXPR
-	    && !bitmap_intersect_p (STORED_SYMS (stmt), variables_loaded))
+	if (is_gimple_assign (stmt)
+	    && AGGREGATE_TYPE_P (TREE_TYPE (gimple_assign_lhs (stmt)))
+	    && (size = lang_hooks.expr_size (gimple_assign_lhs (stmt)))
+	    && integer_zerop (size))
+	  {
+	    if (dump_file && (dump_flags & TDF_DETAILS))
+	      {
+		fprintf (dump_file, "  Deleted zero-sized store '");
+		print_gimple_stmt (dump_file, stmt, 0, dump_flags);
+		fprintf (dump_file, "'\n");
+	      }
+	    removed = true;
+	    gsi_remove (&gsi, true);
+	    todo |= TODO_cleanup_cfg;
+	  }
+	else if (gimple_stored_syms (stmt)
+		 && !bitmap_empty_p (gimple_stored_syms (stmt))
+		 && (is_gimple_assign (stmt)
+		     || (is_gimple_call (stmt)
+			 && gimple_call_lhs (stmt)))
+		 && !bitmap_intersect_p (gimple_stored_syms (stmt),
+					 variables_loaded))
 	  {
 	    unsigned int i;
 	    bitmap_iterator bi;
 	    bool dead = true;
 
-
-
 	    /* See if STMT only stores to write-only variables and
 	       verify that there are no volatile operands.  tree-ssa-operands
 	       sets has_volatile_ops flag for all statements involving
@@ -700,7 +697,7 @@ execute_simple_dse (void)
 	       from removing them as dead.  The flag thus has no use for us
 	       and we need to look into all operands.  */
 	      
-	    EXECUTE_IF_SET_IN_BITMAP (STORED_SYMS (stmt), 0, i, bi)
+	    EXECUTE_IF_SET_IN_BITMAP (gimple_stored_syms (stmt), 0, i, bi)
 	      {
 		tree var = referenced_var_lookup (i);
 		if (TREE_ADDRESSABLE (var)
@@ -709,8 +706,8 @@ execute_simple_dse (void)
 		  dead = false;
 	      }
 
-	    if (dead && LOADED_SYMS (stmt))
-	      EXECUTE_IF_SET_IN_BITMAP (LOADED_SYMS (stmt), 0, i, bi)
+	    if (dead && gimple_loaded_syms (stmt))
+	      EXECUTE_IF_SET_IN_BITMAP (gimple_loaded_syms (stmt), 0, i, bi)
 		if (TREE_THIS_VOLATILE (referenced_var_lookup (i)))
 		  dead = false;
 
@@ -719,59 +716,56 @@ execute_simple_dse (void)
 		if (TREE_THIS_VOLATILE (op))
 		  dead = false;
 
-	    /* Look for possible occurence var = indirect_ref (...) where
+	    /* Look for possible occurrence var = indirect_ref (...) where
 	       indirect_ref itself is volatile.  */
 
-	    if (dead && TREE_THIS_VOLATILE (GIMPLE_STMT_OPERAND (stmt, 1)))
+	    if (dead && is_gimple_assign (stmt)
+	        && TREE_THIS_VOLATILE (gimple_assign_rhs1 (stmt)))
 	      dead = false;
 
 	    if (dead)
 	      {
-		tree call = get_call_expr_in (stmt);
-
 		/* When LHS of var = call (); is dead, simplify it into
 		   call (); saving one operand.  */
-		if (TREE_CODE (stmt) == GIMPLE_MODIFY_STMT
-		    && call
-		    && TREE_SIDE_EFFECTS (call))
+                if (is_gimple_call (stmt)
+                    && gimple_has_side_effects (stmt))
 		  {
 		    if (dump_file && (dump_flags & TDF_DETAILS))
 		      {
 			fprintf (dump_file, "Deleted LHS of call: ");
-			print_generic_stmt (dump_file, stmt, TDF_SLIM);
+			print_gimple_stmt (dump_file, stmt, 0, TDF_SLIM);
 			fprintf (dump_file, "\n");
 		      }
-		    push_stmt_changes (bsi_stmt_ptr (bsi));
-		    TREE_BLOCK (call) = TREE_BLOCK (stmt);
-		    bsi_replace (&bsi, call, false);
-		    maybe_clean_or_replace_eh_stmt (stmt, call);
-		    mark_symbols_for_renaming (call);
-		    pop_stmt_changes (bsi_stmt_ptr (bsi));
+		    push_stmt_changes (gsi_stmt_ptr (&gsi));
+                    gimple_call_set_lhs (stmt, NULL);
+		    pop_stmt_changes (gsi_stmt_ptr (&gsi));
 		  }
 		else
 		  {
 		    if (dump_file && (dump_flags & TDF_DETAILS))
 		      {
 			fprintf (dump_file, "  Deleted dead store '");
-			print_generic_expr (dump_file, stmt, dump_flags);
+			print_gimple_stmt (dump_file, stmt, 0, dump_flags);
 			fprintf (dump_file, "'\n");
 		      }
 		    removed = true;
-		    bsi_remove (&bsi, true);
+		    gsi_remove (&gsi, true);
 		    todo |= TODO_cleanup_cfg;
 		  }
 		todo |= TODO_remove_unused_locals | TODO_ggc_collect;
 	      }
 	  }
 	if (!removed)
-	  bsi_next (&bsi);
+	  gsi_next (&gsi);
       }
   BITMAP_FREE (variables_loaded);
   return todo;
 }
 
-struct tree_opt_pass pass_simple_dse =
+struct gimple_opt_pass pass_simple_dse =
 {
+ {
+  GIMPLE_PASS,
   "sdse",				/* name */
   NULL,					/* gate */
   execute_simple_dse,			/* execute */
@@ -783,6 +777,6 @@ struct tree_opt_pass pass_simple_dse =
   0,					/* properties_provided */
   0,					/* properties_destroyed */
   0,					/* todo_flags_start */
-  TODO_dump_func,          	        /* todo_flags_finish */
-  0				        /* letter */
+  TODO_dump_func          	        /* todo_flags_finish */
+ }
 };
